/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"

static d_open_t  nvidia_ctl_open;
static d_close_t nvidia_ctl_close;
static d_ioctl_t nvidia_ctl_ioctl;
static d_poll_t  nvidia_ctl_poll;

static struct cdevsw nvidia_ctl_cdevsw = {
    /* open    */  nvidia_ctl_open,
    /* close */    nvidia_ctl_close,
    /* read */     noread,
    /* write */    nowrite,
    /* ioctl */    nvidia_ctl_ioctl,
    /* poll */     nvidia_ctl_poll,
    /* mmap */     nommap,
    /* strategy */ nostrategy,
    /* name */     "nvidiactl",
    /* maj */      CDEV_MAJOR,
    /* dump */     nodump,
    /* psize */    nopsize,
    /* flags */    D_TRACKCLOSE,
#if __FreeBSD_version < 500000
    /* bmaj */     -1,
#endif
    /* kqfilter */ NULL,
};

static dev_t          nvidia_ctl_cdev;
static nvidia_softc_t nvidia_ctl_sc;
static int            nvidia_count = 0;

int nvidia_ctl_open(
    dev_t dev,
    int oflags,
    int devtype,
    d_thread_t *td
)
{
    int status;
    nv_state_t *nv = &nvidia_ctl_state;

    nv_lock_api(nv);
    status = nvidia_open_ctl();
    nv_unlock_api(nv);

    return status;
}

int nvidia_ctl_close(
    dev_t dev,
    int fflag,
    int devtype,
    d_thread_t *td
)
{
    int status;
    nv_state_t *nv = &nvidia_ctl_state;

    nv_lock_api(nv);
    status = nvidia_close_ctl(dev, td);
    nv_unlock_api(nv);

    return status;
}

int nvidia_ctl_ioctl(
    dev_t dev,
    u_long cmd,
    caddr_t data,
    int fflag,
    d_thread_t *td
)
{
    int status;
    nv_state_t *nv = &nvidia_ctl_state;

    struct nv_ioctl_card_info    *ci;
    struct nv_ioctl_primary_card *pc;

    ci = (struct nv_ioctl_card_info *)    data;
    pc = (struct nv_ioctl_primary_card *) data;

    nv_lock_api(nv);

    switch (cmd) {
        case NV_IOCTL_CARD_INFO:
            status = nvidia_get_card_info(ci);
            break;

        case NV_IOCTL_PRIMARY_CARD:
            status = nvidia_set_primary(pc);
            break;

        default:
            status = nvidia_handle_ioctl(dev, cmd, data, fflag, td);
            break;
    }

    nv_unlock_api(nv);
    return status;
}


/*
 * The X driver module adds the ctrl device fd to the XFree86 select list
 * and receives notification if the select status changes. We make use of
 * this mechanism to signal hotkey events; this function is called by the
 * core resource manager whenever a legacy hotkey event is detected.
 */

void nv_set_hotkey_occurred_flag(void)
{
    nv_state_t *nv;
    struct nvidia_softc *sc;

    nv = &nvidia_ctl_state;
    sc = nv->os_state;

    nv_lock_rm(nv);
    nv->flags |= NV_FLAG_HOTKEY_OCCURRED;
    nv_unlock_rm(nv);

    selwakeup(&sc->rsel);
}

int nvidia_ctl_poll(
    dev_t dev,
    int events,
    d_thread_t *td
)
{
    nv_state_t *nv;
    struct nvidia_softc *sc;

    nv = &nvidia_ctl_state;
    sc = nv->os_state;

    nv_lock_rm(nv);

    if (nv->flags & NV_FLAG_HOTKEY_OCCURRED) {
        /*
         * A hotkey event has occured. Clear the flag and let X know what
         * happened.
         */
        nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;
        nv_unlock_rm(nv);

        return (POLLIN | POLLRDNORM);
    } else {
        /*
         * There are currently no events pending - we'll have to put this
         * process to sleep until there are.
         */
        nv_unlock_rm(nv);
        selrecord(td, &sc->rsel);
    }

    return 0;
}

int nvidia_ctl_attach(void)
{
    struct nvidia_softc *sc;

    if (nvidia_count == 0) {
        /*
         * This routine is called from nvidia_attach, multiple times when
         * more than one device is installed - this makes it necessary to
         * prevent multiple calls to make_dev.
         */
        nvidia_ctl_cdev = make_dev(&nvidia_ctl_cdevsw,
                CDEV_CTL_MINOR,
                UID_ROOT, GID_WHEEL, 0666,
                "%s", nvidia_ctl_cdevsw.d_name);

        sc = &nvidia_ctl_sc;
        bzero(sc, sizeof(nvidia_softc_t));
        nvidia_ctl_state.os_state = sc;

#if __FreeBSD_version >= 500000
        mtx_init(&sc->mtx_api, "ctl.mtx_api", NULL, MTX_DEF);
        mtx_init(&sc->mtx_rm,  "ctl.mtx_rm",  NULL, MTX_SPIN);
#else
        lockinit(&sc->api_lock, PZERO, "ctl.api_lock", 0, 0); 
#endif
    }

    nvidia_count++;
    return 0;
}

int nvidia_ctl_detach(void)
{
    struct nvidia_softc *sc;

    sc = &nvidia_ctl_sc;
    nvidia_count--;

    if (nvidia_count == 0) {
        /*
         * Similar to nvidia_ctl_attach, nvidia_ctl_detach has to respect
         * the device reference count to prevent premature destruction of
         * the control device.
         */
#if __FreeBSD_version >= 500000
        mtx_destroy(&sc->mtx_rm);
        mtx_destroy(&sc->mtx_api);
#else
        lockmgr(&sc->api_lock, LK_DRAIN, 0, curproc);
#endif
        destroy_dev(nvidia_ctl_cdev);
    }

    return 0;
}

