/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"


/*
 * The NVIDIA kernel module's malloc identifier, needed for both tracking
 * and actual allocation/freeing purposes. M_NVIDIA is declared elsewhere
 * to make it known to other parts of the kernel module (nv-freebsd.h).
 */

MALLOC_DEFINE(M_NVIDIA, "nvidia", "NVIDIA memory allocations");

/*
 * The global client information array; it is used in the open and in the
 * closed source sections of the driver module. The core resource manager
 * was compiled for NV_MAX_CLIENTS also, modifying this define will yield
 * undefined results.
 */

nv_client_t nv_clients[NV_MAX_CLIENTS];


#define MAX_ERROR_STRING    256


RM_STATUS os_alloc_contig_pages(
    VOID **address,
    U032 size
)
{
    *address = contigmalloc(size, M_NVIDIA, 0, 0, ~0, PAGE_SIZE, 0);
    return *address ? RM_OK : RM_ERROR;
}

VOID os_free_contig_pages(
    VOID *address,
    U032 size
)
{
    contigfree(address, size, M_NVIDIA);
}

VOID* os_map_user_space(
    VOID *address,
    VOID **private,
    U032 size,
    U032 mode
)
{
    /*
     * XXX I don't think that there's a good way (any?) to implement this
     * function on FreeBSD. It is not currently used except by the NVIDIA
     * AGP mapping code, which we do not use on FreeBSD, but if it should
     * ever be called from elsewhere, we need to know.
     */
    os_dbg_breakpoint();
    return NULL;
}

VOID os_unmap_user_space(
    VOID *address,
    VOID *private
)
{
    /*
     * XXX I don't think that there's a good way (any?) to implement this
     * function on FreeBSD. It is not currently used except by the NVIDIA
     * AGP mapping code, which we do not use on FreeBSD, but if it should
     * ever be called from elsewhere, we need to know.
     */
    os_dbg_breakpoint();
}

/*
 * The core resource manager's favorite source of memory, this routine is
 * called from different contexts, including ISRs. This means that it can
 * not be allowed to sleep when memory is low.
 */

RM_STATUS os_alloc_mem(
    VOID **address,
    U032 size
)
{
    *address = malloc(size, M_NVIDIA, M_NOWAIT|M_ZERO);
    return *address ? RM_OK : RM_ERROR;
}

VOID os_free_mem(VOID *address)
{
    free(address, M_NVIDIA);
}


RM_STATUS os_delay(U032 MilliSeconds)
{
    DELAY(MilliSeconds * 1000);
    return RM_OK;
}

RM_STATUS os_delay_us(U032 MicroSeconds)
{
    DELAY(MicroSeconds);
    return RM_OK;
}

U032 os_get_cpu_frequency(VOID)
{
    /* XXX x86 only! */
    return ((tsc_freq + 4999) / 1000000);
}

RM_STATUS os_get_current_process(U032 *pPid)
{
    *pPid = curproc->p_pid;
    return RM_OK;
}

RM_STATUS os_kill_process(
    U032 pid,
    U032 sig
)
{
    struct proc *p = pfind(pid);
    if (!p) {
        /*
         * No process with the specified PID was found running. We notify
         * the caller, but take no other action.
         */
        return RM_ERR_OPERATING_SYSTEM;
    }

    psignal(p, sig);

    return RM_OK;
}

RM_STATUS os_get_current_time(
    U032 *sec,
    U032 *usec
)
{
    struct timeval tv;

    getmicrotime(&tv);

    *sec  = tv.tv_sec;
    *usec = tv.tv_usec;

    return RM_OK;
}

BOOL os_is_administrator(PHWINFO pDev)
{
#if __FreeBSD_version < 500000
    return suser(curproc);
#else
    return suser(curthread);
#endif
}

U008 os_io_read_byte(
    PHWINFO pDev,
    U032 address
)
{
    /*
     * XXX This function is not currently implemented, but could be using
     * bus_space_read_1. What is it needed for?
     */
    os_dbg_breakpoint();
    return 0;
}

VOID os_io_write_byte(
    PHWINFO pDev,
    U032 address,
    U008 value
)
{
    /*
     * XXX This function is not currently implemented, but could be using
     * bus_space_write_1. What is it needed for?
     */
    os_dbg_breakpoint();
}

U016 os_io_read_word(
    PHWINFO pDev,
    U032 address
)
{
    /*
     * XXX This function is not currently implemented, but could be using
     * bus_space_read_2. What is it needed for?
     */
    os_dbg_breakpoint();
    return 0;
}

VOID os_io_write_word(
    PHWINFO pDev,
    U032 address,
    U016 value
)
{
    /*
     * XXX This function is not currently implemented, but could be using
     * bus_space_write_2. What is it needed for?
     */
    os_dbg_breakpoint();
}

U032 os_io_read_dword(
    PHWINFO pDev,
    U032 address
)
{
    /*
     * XXX This function is not currently implemented, but could be using
     * bus_space_read_4. What is it needed for?
     */
    os_dbg_breakpoint();
    return 0;
}

VOID os_io_write_dword(
    PHWINFO pDev,
    U032 address,
    U032 value
)
{
    /*
     * XXX This function is not currently implemented, but could be using
     * bus_space_write_4. What is it needed for?
     */
    os_dbg_breakpoint();
}

VOID* os_map_io_space(
    U032 start,
    U032 size,
    VOID **private,
    U032 user,
    U032 mode
)
{
    os_dbg_breakpoint();
    return NULL;
}

VOID os_unmap_io_space(
    VOID *address,
    U032 size,
    VOID *private,
    U032 user
)
{
    /*
     * XXX While os_map_io_space does not appear to be called, this unmap
     * routine apparantly still is.
     */
}

/*
 * This routine maps a range of physical pages into kernel virtual memory
 * to make it accessible by the kernel module. Normally, we wouldn't need
 * this, the kernel provides drivers with linear mappings for each of the
 * memory resources of the device the driver registered itself for.
 *
 * One of the aspects of this driver that makes this routine necessary is
 * the built-in AGP GART driver.
 */

VOID* os_map_kernel_space(
    U032 start,
    U032 size,
    U032 mode
)
{
    void *vm;
    vm = pmap_mapdev(start, size);

    if (vm != NULL && mode == NV_MEMORY_UNCACHED) {
        unsigned *pte;
        vm_offset_t va = (vm_offset_t) vm;
        do {
            pte = (unsigned *) vtopte(va);
            *pte |= PG_N;
            va += PAGE_SIZE;
        } while (va < (vm_offset_t) vm + size);
        invltlb();
    }

    return vm;
}

VOID os_unmap_kernel_space(
    VOID *address,
    U032 size
)
{
    /* delete kmap and pmap */
    pmap_unmapdev((vm_offset_t) address, size);
}

RM_STATUS os_set_mem_range(
    U032 start,
    U032 size,
    U032 mode
)
{
    int arg;
    struct mem_range_desc mrd;

    mrd.mr_base  = start;
    mrd.mr_len   = size;
    mrd.mr_flags = MDF_WRITECOMBINE;

    strcpy(mrd.mr_owner, "NVIDIA");
    arg = MEMRANGE_SET_UPDATE;

    if (mem_range_attr_set(&mrd, &arg))
        return RM_ERROR;

    return RM_OK;
}

RM_STATUS os_unset_mem_range(
    U032 start,
    U032 size
)
{
    int arg;
    struct mem_range_desc mrd;

    mrd.mr_base = start;
    mrd.mr_len  = size;
    arg = MEMRANGE_SET_REMOVE;

    if (mem_range_attr_set(&mrd, &arg))
        return RM_ERROR;

    return RM_OK;
}


ULONG os_sti(ULONG flags)
{
    /*
     * XXX This routine is no longer used and should probably be removed.
     */
    return flags;
}

ULONG os_cli(ULONG flags)
{
    /*
     * XXX This routine is no longer used and should probably be removed.
     */
    return flags;
}


/*
 *
 * void out_string (const char *str)
 * {
 *     printf("%s", str);
 * }
 *
 */

/*
 * The current debug level is used to determine if certain debug messages
 * are printed to the system console/log files or not. It defaults to the
 * highest debug level, i.e. the lowest debug output.
 */

int cur_debuglevel = 0xaaaaaaaa;

VOID os_dbg_init(VOID)
{
    U032 new_debuglevel;

    if (os_read_registry_dword(NULL, NULL,
                               "ResmanDebugLevel",
                               &new_debuglevel) == RM_OK)
    {
        if (new_debuglevel != ~0)
            cur_debuglevel = new_debuglevel;
    }
}

VOID os_dbg_breakpoint(VOID)
{
#ifdef DEBUG
    /*
     * Enter the FreeBSD DDB kernel debugger; this interface is valid for
     * all debugger frontends this kernel may support. Of interest for us
     * is the GDB stub, which allows source-level remote debugging.
     */
    Debugger("os_dbg_breakpoint()");
#endif
}


int
nv_printf (const char *fmt, ...)
{
    char    nv_error_string[MAX_ERROR_STRING];
    char    *p;
    int     chars_written;
    va_list arglist;

    p = nv_error_string;

    va_start (arglist, fmt);
    chars_written = vsprintf (p, fmt, arglist);
    va_end (arglist);

    printf("%s", p);

    return (chars_written);
}


S032 os_mem_cmp(
    const U008 *buf0,
    const U008 *buf1,
    U032 length
)
{
    return memcmp(buf0, buf1, length);
}

U008* os_mem_copy(
    U008 *dst,
    const U008 *src,
    U032 length
)
{
    return memcpy(dst, src, length);
}

RM_STATUS os_memcpy_from_user(
    VOID *dst,
    const VOID *src,
    U032 length
)
{
    if (src < (void *) VM_MAXUSER_ADDRESS)
        return copyin(src, dst, length)  ? RM_ERR_BAD_ADDRESS : RM_OK;

    return os_mem_copy(dst, src, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

RM_STATUS os_memcpy_to_user(
    VOID *dst,
    const VOID *src,
    U032 length
)
{
    if (dst < (void *) VM_MAXUSER_ADDRESS)
        return copyout(src, dst, length) ? RM_ERR_BAD_ADDRESS : RM_OK;

    return os_mem_copy(dst, src, length) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

VOID* os_mem_set(
    VOID *b,
    U008 c,
    U032 length
)
{
    return memset(b, c, length);
}

S032 os_string_compare(
    const U008 *s1,
    const U008 *s2
)
{
    return strcmp(s1, s2);
}

U008* os_string_copy(
    U008 *dst,
    const U008 *src
)
{
    return strcpy(dst, src);
}

U032 os_string_length(const U008* s)
{
    return strlen(s);
}

RM_STATUS os_strncpy_from_user(
    U008 *dst,
    const U008 *src,
    U032 n
)
{
    int done;

    return copyinstr(src, dst, n, &done) ? RM_ERR_BAD_ADDRESS : RM_OK;
}

VOID* os_copy_in_ioctl_param(
    VOID *dst,
    VOID *src,
    U032 length
)
{
    return (src);
}

VOID* os_copy_out_ioctl_param(
    VOID *dst,
    VOID *src,
    U032 length
)
{
    return (void *)1;
}


U032 os_get_page_size(VOID)
{
    return PAGE_SIZE;
}

U032 os_get_page_mask(VOID)
{
    /*
     * On FreeBSD, PAGE_MASK means (PAGE_SIZE - 1); on Linux it means the
     * opposite, ~(PAGE_SIZE - 1); that is what this function is expected
     * to return.
     */
    return ~PAGE_MASK;
}


RM_STATUS os_init_swap_barrier(U032 id)
{
    /*
     * XXX Support for the SGI swap barrier mechanism. I don't think this
     * is actually supported on FreeBSD by SGI itself, maybe the calls to
     * the related functions should be made Linux specific?
     */
    return RM_ERROR;
}

RM_STATUS os_remove_swap_barrier(U032 id)
{
    /*
     * XXX Support for the SGI swap barrier mechanism. I don't think this
     * is actually supported on FreeBSD by SGI itself, maybe the calls to
     * the related functions should be made Linux specific?
     */
    return RM_ERROR;
}

RM_STATUS os_swap_barrier(
    U032 id,
    VOID (*callback)(VOID *),
    VOID *data
)
{
    /*
     * XXX Support for the SGI swap barrier mechanism. I don't think this
     * is actually supported on FreeBSD by SGI itself, maybe the calls to
     * the related functions should be made Linux specific?
     */
    return RM_ERROR;
}

