/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"
#include "nv-ids.h"

#define NVIDIA_VENDORID 0x10de
#define NVIDIA_PCI_REG  0x10
#define NVIDIA_PCI_MEM  0x14

static int nvidia_pci_probe  (device_t);
static int nvidia_pci_attach (device_t);
static int nvidia_pci_detach (device_t);

int nvidia_pci_probe(device_t dev)
{
    int i;
    int devices = sizeof(sChipInfoChar) / sizeof(CHIP_INFO_CHAR);

    if (pci_get_vendor(dev) != NVIDIA_VENDORID)
        return ENXIO;

    for (i = 0; i < devices; i++) {
        if (pci_get_device(dev) == sChipInfoChar[i].ulDevID) {
            /* device_quiet(dev); */
            device_set_desc(dev, sChipInfoChar[i].psChip);
            return 0;
        }
    }

    return ENXIO;
}

int nvidia_pci_attach(device_t dev)
{
    int status;
    struct nvidia_softc *sc;
    U016 word;

    sc = device_get_softc(dev);

    pci_enable_busmaster(dev);
    word = pci_read_config(dev, PCIR_COMMAND, 2);

    if ((word & PCIM_CMD_BUSMASTEREN) == 0) {
        device_printf(dev, "Unable to enable PCI busmastering.\n");
        status = ENXIO;
        goto fail;
    }

    pci_enable_io(dev, SYS_RES_MEMORY);
    word = pci_read_config(dev, PCIR_COMMAND, 2);

    if ((word & PCIM_CMD_MEMEN) == 0) {
        device_printf(dev, "Unable to enable PCI memory resources.\n");
        status = ENXIO;
        goto fail;
    }

    sc->reg_type = SYS_RES_MEMORY;
    sc->reg_rid  = NVIDIA_PCI_REG;
    sc->mem_type = SYS_RES_MEMORY;
    sc->mem_rid  = NVIDIA_PCI_MEM;
    sc->irq_rid  = 0;

    status = nvidia_alloc(dev);
    if (status) {
        goto fail;
    }

    status = nvidia_attach(dev);
    if (status) {
        device_printf(dev, "NVIDIA driver attach failed.\n");
        goto fail;
    }

    status = bus_setup_intr(dev, sc->irq, INTR_TYPE_MISC,
                                nvidia_intr, sc, &sc->irq_ih);
    if (status) {
        device_printf(dev, "Interrupt handler setup failed.\n");
        goto fail;
    }

#if __FreeBSD_version >= 500000
    mtx_init(&sc->mtx_api, "dev.mtx_api", NULL, MTX_DEF);
    mtx_init(&sc->mtx_rm,  "dev.mtx_rm",  NULL, MTX_SPIN);
#else
    lockinit(&sc->api_lock, PZERO, "dev.api_lock", 0, 0); 
#endif
    return 0;

fail:
    nvidia_free(dev);
    return status;
}

int nvidia_pci_detach(device_t dev)
{
    int status;
    struct nvidia_softc *sc;

    sc = device_get_softc(dev);
#if __FreeBSD_version >= 500000
    mtx_destroy(&sc->mtx_rm);
    mtx_destroy(&sc->mtx_api);
#else
    lockmgr(&sc->api_lock, LK_DRAIN, 0, curproc);
#endif

    status = nvidia_detach(dev);
    if (status)
        goto fail;

    status = nvidia_free(dev);
    if (status)
        goto fail;

fail:
    return status;
}

static device_method_t nvidia_pci_methods[] = {
    DEVMETHOD( device_probe,   nvidia_pci_probe  ),
    DEVMETHOD( device_attach,  nvidia_pci_attach ),
    DEVMETHOD( device_detach,  nvidia_pci_detach ),
#if 0
    DEVMETHOD( device_suspend, nvidia_suspend    ),
    DEVMETHOD( device_resume,  nvidia_resume     ),
#endif
    { 0, 0 }
};

static driver_t nvidia_pci_driver = {
    "nvidia",
    nvidia_pci_methods,
    sizeof(struct nvidia_softc)
};

DRIVER_MODULE(nvidia, pci, nvidia_pci_driver, nvidia_devclass, nvidia_modevent, 0);

#ifdef USE_OS_AGP_GART 
MODULE_DEPEND(nvidia, agp, 1, 1, 1);
#endif

#ifdef NV_SUPPORT_LINUX_COMPAT
MODULE_DEPEND(nvidia, linux, 1, 1, 1);
#endif

