/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include "nv-misc.h"
#include "os-interface.h"
#include "nv.h"
#include "nv-freebsd.h"

struct sysctl_ctx_list sysctl_ctx;

struct sysctl_oid *oid_nvidia;
struct sysctl_oid *oid_agp;
struct sysctl_oid *oid_registry;

void nvidia_sysctl_init(void)
{
    device_t dev;
    struct sysctl_oid *oid;

    U032 i = 0;
    struct nvidia_registry *entry;

    sysctl_ctx_init(&sysctl_ctx);

    oid_nvidia = SYSCTL_ADD_NODE(&sysctl_ctx,
            SYSCTL_STATIC_CHILDREN(_hw),
            OID_AUTO,
            "nvidia",
            CTLFLAG_RD | CTLFLAG_DYN,
            0,
            "NVIDIA SYSCTL Master Node");

    oid_agp = SYSCTL_ADD_NODE(&sysctl_ctx,
            SYSCTL_CHILDREN(oid_nvidia),
            OID_AUTO,
            "agp",
            CTLFLAG_RD | CTLFLAG_DYN,
            0,
            "NVIDIA SYSCTL AGP Node");

    SYSCTL_ADD_STRING(&sysctl_ctx,
            SYSCTL_CHILDREN(oid_nvidia),
            OID_AUTO,
            "version",
            CTLFLAG_RD | CTLFLAG_DYN,
            (char *)(uintptr_t) pNVRM_ID,
            0,
            "NVIDIA Resource Manager (NVRM) Version");


    oid_registry = SYSCTL_ADD_NODE(&sysctl_ctx,
            SYSCTL_CHILDREN(oid_nvidia),
            OID_AUTO,
            "registry",
            CTLFLAG_RD | CTLFLAG_DYN,
            0,
            "NVIDIA SYSCTL Registry Node");

    do {
        entry = &nvreg[i++];
        if (entry->valid != 0) {
            /*
             * This entry is valid, i.e. processed by the registry access
             * routines; add it to our SYSCTL registry interface.
             */
            SYSCTL_ADD_UINT(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid_registry),
                    OID_AUTO,
                    entry->name,
                    CTLFLAG_RW | CTLFLAG_DYN,
                    &entry->data, 0,
                    entry->description);
        }
    } while(entry->name != NULL);


    dev = nvidia_find_bridge();
    if (!dev) {
        /*
         * We were unable to locate an AGP capable host bridge; depending
         * on whether or not this system is equipped with an AGP port, we
         * may have been looking in the wrong place. Either way, we can't
         * report any information about it.
         */
        return;
    }

    oid = SYSCTL_ADD_NODE(&sysctl_ctx,
            SYSCTL_CHILDREN(oid_agp),
            OID_AUTO,
            "host-bridge",
            CTLFLAG_RD | CTLFLAG_DYN,
            0,
            "NVIDIA AGP Device Node");

    SYSCTL_ADD_PROC(&sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "rates",
            CTLFLAG_RD | CTLFLAG_DYN,
            (void *) dev, 0,
            nvidia_sysctl_rates,
            "A",
            "NVIDIA AGP Device Rate Info");

    SYSCTL_ADD_PROC(&sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "fw",
            CTLFLAG_RD | CTLFLAG_DYN,
            (void *) dev, 0,
            nvidia_sysctl_fw,
            "A",
            "NVIDIA AGP Device FW Info");

    SYSCTL_ADD_PROC(&sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "sba",
            CTLFLAG_RD | CTLFLAG_DYN,
            (void *) dev, 0,
            nvidia_sysctl_sba,
            "A",
            "NVIDIA AGP Device SBA Info");

    SYSCTL_ADD_PROC(&sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "registers",
            CTLFLAG_RD | CTLFLAG_DYN,
            (void *) dev, 0,
            nvidia_sysctl_regs,
            "A",
            "NVIDIA AGP Device Registers");
}

void nvidia_sysctl_exit(void)
{
    sysctl_ctx_free(&sysctl_ctx);
}


U008 nvidia_find_cap(device_t dev)
{
    U016 status;
    U008 cap_ptr, cap_id;

    status = pci_read_config(dev, PCIR_STATUS, 2);
    status &= PCIM_STATUS_CAPPRESENT;

    if (!status) {
        /*
         * A PCI device without capability list? Hardly possible with any
         * NVIDIA card, but better safe than sorry...
         */
        return 0;
    }

    switch (pci_get_class(dev)) {
        case PCIC_DISPLAY:
        case PCIC_BRIDGE:
            cap_ptr = pci_read_config(dev, PCIR_CAP_PTR, 1);
            break;
        default:
            return 0;
    }

    do {
        cap_ptr &= 0xfc;
        cap_id = pci_read_config(dev, cap_ptr + PCIR_CAP_LIST_ID, 1);
        if (cap_id == PCIR_CAP_ID_AGP) {
            /*
             * We are in luck, traversing this device's capabilities list
             * produced one of type AGP. This is what we came for, return
             * the capability pointer.
             */
            return cap_ptr;
        }
        cap_ptr = pci_read_config(dev, cap_ptr + PCIR_CAP_LIST_NEXT, 1);
    } while (cap_ptr && cap_id != 0xff);

    return 0;
}

void* nvidia_find_bridge(void)
{
    U008 hdrtype;
    U032 bus;
    U032 device;
    U032 func;

    device_t dev;

    for (bus = 0; bus < 5; bus++) {
        /*
         * Attempt to find the host bridge with brute force, there are no
         * routines to find a PCI device by class at this point. In order
         * to locate the bridge, we iterate over reasonable bus, slot and
         * device function combinations.
         */
        for (device = 0; device < 32; device++) {
            for (func = 0; func < 8; func++) {
                /*
                 * We're not interested in non-bridge PCI devices, nor do
                 * care for devices with our own or an invalid vendor ID.
                 */
                dev = pci_find_bsf(bus, device, func);

                if (dev == NULL)
                    break;

                if (pci_get_class(dev) != PCIC_BRIDGE)
                    break;

                if (pci_get_vendor(dev) == 0x10de)
                    break;

                if (pci_get_vendor(dev) == 0xffff)
                    break;

                if (nvidia_find_cap(dev) != 0)
                    return dev;

                /*
                 * It also doesn't make sense to to iterate over multiple
                 * functions if this isn't a multi-function device.
                 */
                hdrtype = pci_read_config(dev, PCIR_HEADERTYPE, 1);

                if ((hdrtype & PCIM_MFDEV) == 0)
                    break;
            }
        }
    }

    return NULL;
}


int nvidia_sysctl_vbios(SYSCTL_HANDLER_ARGS)
{
    nv_state_t *nv = arg1;

    U032 vbios[5];
    U008 vbios_version[16];

    if (rm_get_vbios_version(nv, &vbios[0], &vbios[1], &vbios[2],
                &vbios[3], &vbios[4]) != RM_OK) {
        /*
         * The VBIOS version is only accessible after the device has been
         * initialized with rm_init_adapter. 
         */
        sprintf(vbios_version, "??.??.??.??.??");
    } else {
        sprintf(vbios_version, "%02x.%02x.%02x.%02x.%02x", vbios[0],
                vbios[1], vbios[2], vbios[3], vbios[4]);
    }

    return SYSCTL_OUT(req, vbios_version, strlen(vbios_version) + 1);
}

int nvidia_sysctl_type(SYSCTL_HANDLER_ARGS)
{
    U008 type[4];
    struct nvidia_softc *sc = arg1;

    if (nvidia_find_cap(sc->dev) != 0) {
        /*
         * We found an AGP capability pointer in this device's PCI config
         * space, let's inform the user that he/she is the proud owner of
         * an AGP card.
         */
        sprintf(type, "AGP");
    } else {
        sprintf(type, "PCI");
    }

    return SYSCTL_OUT(req, type, strlen(type) + 1);
}

int nvidia_sysctl_rates(SYSCTL_HANDLER_ARGS)
{
    device_t dev = arg1;

    U032 status;
    U008 cap_ptr, rate[16];

    cap_ptr = nvidia_find_cap(dev);
    status  = pci_read_config(dev, cap_ptr + 4, 4);

    sprintf(rate, "%s%s%s",
            (status & 0x0004) ? "4x " : "",
            (status & 0x0002) ? "2x " : "",
            (status & 0x0001) ? "1x " : "");

    return SYSCTL_OUT(req, rate, strlen(rate) + 1);
}

int nvidia_sysctl_fw(SYSCTL_HANDLER_ARGS)
{
    device_t dev = arg1;

    U032 status;
    U008 cap_ptr, fw[16];

    cap_ptr = nvidia_find_cap(dev);
    status  = pci_read_config(dev, cap_ptr + 4, 4);

    sprintf(fw, "%s",
            (status & 0x0010) ? "supported" : "not supported");

    return SYSCTL_OUT(req, fw, strlen(fw) + 1);
}

int nvidia_sysctl_sba(SYSCTL_HANDLER_ARGS)
{
    device_t dev = arg1;

    U032 status;
    U008 cap_ptr, sba[16];

    cap_ptr = nvidia_find_cap(dev);
    status  = pci_read_config(dev, cap_ptr + 4, 4);

    sprintf(sba, "%s",
            (status & 0x0200) ? "supported" : "not supported");

    return SYSCTL_OUT(req, sba, strlen(sba) + 1);
}


int nvidia_sysctl_regs(SYSCTL_HANDLER_ARGS)
{
    device_t dev = arg1;

    U032 command;
    U032 status;
    U008 cap_ptr, registers[24];

    cap_ptr = nvidia_find_cap(dev);
    status  = pci_read_config(dev, cap_ptr + 4, 4);
    command = pci_read_config(dev, cap_ptr + 8, 4);

    sprintf(registers, "0x%08x:0x%08x", status, command);

    return SYSCTL_OUT(req, registers, strlen(registers) + 1);
}

int nvidia_sysctl_driver(SYSCTL_HANDLER_ARGS)
{
    U008 driver[32];
    nv_state_t *nv = arg1;

    switch (nv->agp_config) {
        case NVOS_AGP_CONFIG_DISABLE_AGP:
            sprintf(driver, "n/a");
            break;
        case NVOS_AGP_CONFIG_OSAGP:
            sprintf(driver, "freebsd (agp.ko)");
            break;
        case NVOS_AGP_CONFIG_NVAGP:
            sprintf(driver, "nvidia");
            break;
    }

    return SYSCTL_OUT(req, driver, strlen(driver) + 1);
}

int nvidia_sysctl_rate_s(SYSCTL_HANDLER_ARGS)
{
    device_t bridge, card = arg1;
    
    U032 command;
    U032 scratch;
    U008 cap_ptr, rate[4];

    cap_ptr = nvidia_find_cap(card);
    command = pci_read_config(card, cap_ptr + 8, 4);

    bridge  = nvidia_find_bridge();
    cap_ptr = nvidia_find_cap(bridge);

    scratch = pci_read_config(bridge, cap_ptr + 8, 4);
    command &= scratch;

    if ((command & 0x100) != 0) {
        /*
         * Providing AGP status information only makes sense if both card
         * and host-bridge indicate that they are in AGP mode.
         */
        sprintf(rate, "%1dx", command & 0x0007);
    } else {
        sprintf(rate, "n/a");
    }

    return SYSCTL_OUT(req, rate, strlen(rate) + 1);
}

int nvidia_sysctl_fw_s(SYSCTL_HANDLER_ARGS)
{
    device_t bridge, card = arg1;

    U032 command;
    U032 scratch;
    U008 cap_ptr, fw[16];

    cap_ptr = nvidia_find_cap(card);
    command = pci_read_config(card, cap_ptr + 8, 4);

    bridge  = nvidia_find_bridge();
    cap_ptr = nvidia_find_cap(bridge);

    scratch = pci_read_config(bridge, cap_ptr + 8, 4);
    command &= scratch;

    if ((command & 0x0100) != 0) {
        /*
         * Providing AGP status information only makes sense if both card
         * and host-bridge indicate that the are in AGP mode.
         */
        sprintf(fw, "%s",
                (command & 0x0010) ? "enabled" : "disabled");
    } else {
        sprintf(fw, "n/a");
    }

    return SYSCTL_OUT(req, fw, strlen(fw) + 1);
}
    
int nvidia_sysctl_sba_s(SYSCTL_HANDLER_ARGS)
{
    device_t bridge, card = arg1;

    U032 command;
    U032 scratch;
    U008 cap_ptr, sba[16];

    cap_ptr = nvidia_find_cap(card);
    command = pci_read_config(card, cap_ptr + 8, 4);

    bridge  = nvidia_find_bridge();
    cap_ptr = nvidia_find_cap(bridge);

    scratch = pci_read_config(bridge, cap_ptr + 8, 4);
    command &= scratch;

    if ((command & 0x0100) != 0) {
        /*
         * Providing AGP status information only makes sense if both card
         * and host-bridge indicate that the are in AGP mode.
         */
        sprintf(sba, "%s",
                (command & 0x0200) ? "enabled" : "disabled");
    } else {
        sprintf(sba, "n/a");
    }

    return SYSCTL_OUT(req, sba, strlen(sba) + 1);
}
    
int nvidia_sysctl_status(SYSCTL_HANDLER_ARGS)
{
    device_t bridge, card = arg1;

    U032 command;
    U032 scratch;
    U008 cap_ptr, status[16];

    cap_ptr = nvidia_find_cap(card);
    command = pci_read_config(card, cap_ptr + 8, 4);

    bridge  = nvidia_find_bridge();
    cap_ptr = nvidia_find_cap(bridge);

    scratch = pci_read_config(bridge, cap_ptr + 8, 4);
    command &= scratch;

    sprintf(status, "%s",
            (command & 0x0100) ? "enabled" : "disabled");

    return SYSCTL_OUT(req, status, strlen(status) + 1);
}
    

void nv_sysctl_init(nv_state_t *nv)
{
    struct sysctl_oid *oid;
    struct nvidia_softc *sc = nv->os_state;

    char name[4];
    sprintf(name, "%d", device_get_unit(sc->dev));

    sysctl_ctx_init(&sc->sysctl_ctx);

    oid = SYSCTL_ADD_NODE(&sc->sysctl_ctx,
            SYSCTL_CHILDREN(oid_nvidia),
            OID_AUTO,
            "cards",
            CTLFLAG_RD | CTLFLAG_DYN,
            0,
            "NVIDIA SYSCTL Cards Node");

    oid = SYSCTL_ADD_NODE(&sc->sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            name,
            CTLFLAG_RD | CTLFLAG_DYN,
            0,
            "NVIDIA SYSCTL Device Node");

    SYSCTL_ADD_STRING(&sc->sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "model",
            CTLFLAG_RD | CTLFLAG_DYN,
            (char *)(uintptr_t) device_get_desc(sc->dev),
            0,
            "NVIDIA Device Model Name");

    SYSCTL_ADD_UINT(&sc->sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "irq",
            CTLFLAG_RD | CTLFLAG_DYN,
            &nv->interrupt_line,
            0,
            "NVIDIA Device IRQ Number");

    SYSCTL_ADD_PROC(&sc->sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "vbios",
            CTLFLAG_RD | CTLFLAG_DYN,
            (void *) nv, 0,
            nvidia_sysctl_vbios,
            "A",
            "NVIDIA Device VBIOS Version");

    SYSCTL_ADD_PROC(&sc->sysctl_ctx,
            SYSCTL_CHILDREN(oid),
            OID_AUTO,
            "type",
            CTLFLAG_RD | CTLFLAG_DYN,
            (void *) sc, 0,
            nvidia_sysctl_type,
            "A",
            "NVIDIA Device Bus Type");


    if (nvidia_find_cap(sc->dev) != 0) {
        /*
         * This is an AGP device, the only one possible in a given system
         * and hence the one we want to monitor with hw.nvidia.agp.card.
         */

        oid = SYSCTL_ADD_NODE(&sysctl_ctx,
                SYSCTL_CHILDREN(oid_agp),
                OID_AUTO,
                "card",
                CTLFLAG_RD | CTLFLAG_DYN,
                0,
                "NVIDIA AGP Device Node");

        SYSCTL_ADD_PROC(&sysctl_ctx,
                SYSCTL_CHILDREN(oid),
                OID_AUTO,
                "rates",
                CTLFLAG_RD | CTLFLAG_DYN,
                (void *) sc->dev, 0,
                nvidia_sysctl_rates,
                "A",
                "NVIDIA AGP Device Rate Info");

        SYSCTL_ADD_PROC(&sysctl_ctx,
                SYSCTL_CHILDREN(oid),
                OID_AUTO,
                "fw",
                CTLFLAG_RD | CTLFLAG_DYN,
                (void *) sc->dev, 0,
                nvidia_sysctl_fw,
                "A",
                "NVIDIA AGP Device FW Info");

        SYSCTL_ADD_PROC(&sysctl_ctx,
                SYSCTL_CHILDREN(oid),
                OID_AUTO,
                "sba",
                CTLFLAG_RD | CTLFLAG_DYN,
                (void *) sc->dev, 0,
                nvidia_sysctl_sba,
                "A",
                "NVIDIA AGP Device SBA Info");

        SYSCTL_ADD_PROC(&sysctl_ctx,
                SYSCTL_CHILDREN(oid),
                OID_AUTO,
                "registers",
                CTLFLAG_RD | CTLFLAG_DYN,
                (void *) sc->dev, 0,
                nvidia_sysctl_regs,
                "A",
                "NVIDIA AGP Device Registers");


        if (nvidia_find_bridge() != NULL) {
            /*
             * The AGP capabilities of card and host-bridge are displayed
             * separately; the following oids merge the data from both in
             * an attempt to display accurate AGP status information.
             */

            oid = SYSCTL_ADD_NODE(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid_agp),
                    OID_AUTO,
                    "status",
                    CTLFLAG_RD | CTLFLAG_DYN,
                    0,
                    "NVIDIA AGP Status Node");

            SYSCTL_ADD_PROC(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid),
                    OID_AUTO,
                    "status",
                    CTLFLAG_RD | CTLFLAG_DYN,
                    (void *) sc->dev, 0,
                    nvidia_sysctl_status,
                    "A",
                    "NVIDIA AGP Status Information");

            SYSCTL_ADD_PROC(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid),
                    OID_AUTO,
                    "driver",
                    CTLFLAG_RD | CTLFLAG_DYN,
                    (void *) nv, 0,
                    nvidia_sysctl_driver,
                    "A",
                    "NVIDIA AGP Driver Information");

            SYSCTL_ADD_PROC(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid),
                    OID_AUTO,
                    "rate",
                    CTLFLAG_RD | CTLFLAG_DYN,
                    (void *) sc->dev, 0,
                    nvidia_sysctl_rate_s,
                    "A",
                    "NVIDIA AGP Rate Status Information");

            SYSCTL_ADD_PROC(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid),
                    OID_AUTO,
                    "fw",
                    CTLFLAG_RD | CTLFLAG_DYN,
                    (void *) sc->dev, 0,
                    nvidia_sysctl_fw_s,
                    "A",
                    "NVIDIA AGP FW Status Information");

            SYSCTL_ADD_PROC(&sysctl_ctx,
                    SYSCTL_CHILDREN(oid),
                    OID_AUTO,
                    "sba",
                    CTLFLAG_RD | CTLFLAG_DYN,
                    (void *) sc->dev, 0,
                    nvidia_sysctl_sba_s,
                    "A",
                    "NVIDIA AGP SBA Status Information");
        }
    }
}

void nv_sysctl_exit(nv_state_t *nv)
{
    struct nvidia_softc *sc = nv->os_state;
    sysctl_ctx_free(&sc->sysctl_ctx);
}

